/* foundry-debugger-mapped-region.h
 *
 * Copyright 2025 Christian Hergert <chergert@redhat.com>
 *
 * This library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of the
 * License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * SPDX-License-Identifier: LGPL-2.1-or-later
 */

#pragma once

#include <glib-object.h>

#include "foundry-version-macros.h"

G_BEGIN_DECLS

#define FOUNDRY_TYPE_DEBUGGER_MAPPED_REGION (foundry_debugger_mapped_region_get_type())

FOUNDRY_AVAILABLE_IN_ALL
G_DECLARE_DERIVABLE_TYPE (FoundryDebuggerMappedRegion, foundry_debugger_mapped_region, FOUNDRY, DEBUGGER_MAPPED_REGION, GObject)

struct _FoundryDebuggerMappedRegionClass
{
  GObjectClass parent_class;

  void     (*get_range)  (FoundryDebuggerMappedRegion *self,
                          guint64                     *begin_address,
                          guint64                     *end_address);
  char    *(*dup_path)   (FoundryDebuggerMappedRegion *self);
  guint    (*get_mode)   (FoundryDebuggerMappedRegion *self);
  guint64  (*get_offset) (FoundryDebuggerMappedRegion *self);

  /*< private >*/
  gpointer _reserved[11];
};

FOUNDRY_AVAILABLE_IN_ALL
void   foundry_debugger_mapped_region_get_range    (FoundryDebuggerMappedRegion *self,
                                                    guint64                     *beign_address,
                                                    guint64                     *end_address);
FOUNDRY_AVAILABLE_IN_ALL
char    *foundry_debugger_mapped_region_dup_path   (FoundryDebuggerMappedRegion *self);
FOUNDRY_AVAILABLE_IN_ALL
guint    foundry_debugger_mapped_region_get_mode   (FoundryDebuggerMappedRegion *self);
FOUNDRY_AVAILABLE_IN_1_1
guint64  foundry_debugger_mapped_region_get_offset (FoundryDebuggerMappedRegion *self);

G_END_DECLS
