/* This file is generated by glib-genmarshal, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */
#pragma once

#include <glib-object.h>

G_BEGIN_DECLS

/* VOID:OBJECT,OBJECT (../foundry/libfoundry/foundry-marshal.defs:1) */
G_GNUC_INTERNAL
void foundry_marshal_VOID__OBJECT_OBJECT (GClosure     *closure,
                                          GValue       *return_value,
                                          guint         n_param_values,
                                          const GValue *param_values,
                                          gpointer      invocation_hint,
                                          gpointer      marshal_data);
G_GNUC_INTERNAL
void foundry_marshal_VOID__OBJECT_OBJECTv (GClosure *closure,
                                           GValue   *return_value,
                                           gpointer  instance,
                                           va_list   args,
                                           gpointer  marshal_data,
                                           int       n_params,
                                           GType    *param_types);

/* VOID:STRING (../foundry/libfoundry/foundry-marshal.defs:2) */
#define foundry_marshal_VOID__STRING	g_cclosure_marshal_VOID__STRING
#define foundry_marshal_VOID__STRINGv	g_cclosure_marshal_VOID__STRINGv

/* VOID:VOID (../foundry/libfoundry/foundry-marshal.defs:3) */
#define foundry_marshal_VOID__VOID	g_cclosure_marshal_VOID__VOID
#define foundry_marshal_VOID__VOIDv	g_cclosure_marshal_VOID__VOIDv


G_END_DECLS
