/*
 * This file is generated by gdbus-codegen, do not modify it.
 *
 * The license of this code is the same as for the D-Bus interface description
 * it was derived from. Note that it links to GLib, so must comply with the
 * LGPL linking clauses.
 */

#ifndef __FOUNDRY_IPC_H__
#define __FOUNDRY_IPC_H__

#include <gio/gio.h>

G_BEGIN_DECLS


/* ------------------------------------------------------------------------ */
/* Declarations for org.gnome.foundry.CommandLineService */

#define FOUNDRY_IPC_TYPE_COMMAND_LINE_SERVICE (foundry_ipc_command_line_service_get_type ())
#define FOUNDRY_IPC_COMMAND_LINE_SERVICE(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), FOUNDRY_IPC_TYPE_COMMAND_LINE_SERVICE, FoundryIpcCommandLineService))
#define FOUNDRY_IPC_IS_COMMAND_LINE_SERVICE(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), FOUNDRY_IPC_TYPE_COMMAND_LINE_SERVICE))
#define FOUNDRY_IPC_COMMAND_LINE_SERVICE_GET_IFACE(o) (G_TYPE_INSTANCE_GET_INTERFACE ((o), FOUNDRY_IPC_TYPE_COMMAND_LINE_SERVICE, FoundryIpcCommandLineServiceIface))

struct _FoundryIpcCommandLineService;
typedef struct _FoundryIpcCommandLineService FoundryIpcCommandLineService;
typedef struct _FoundryIpcCommandLineServiceIface FoundryIpcCommandLineServiceIface;

struct _FoundryIpcCommandLineServiceIface
{
  GTypeInterface parent_iface;

  gboolean (*handle_run) (
    FoundryIpcCommandLineService *object,
    GDBusMethodInvocation *invocation,
    GUnixFDList *fd_list,
    const gchar *arg_directory,
    const gchar *const *arg_environment,
    const gchar *const *arg_argv,
    GVariant *arg_stdin_handle,
    GVariant *arg_stdout_handle,
    GVariant *arg_stderr_handle,
    const gchar *arg_object_path);

};

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (FoundryIpcCommandLineService, g_object_unref)
#endif

GType foundry_ipc_command_line_service_get_type (void) G_GNUC_CONST;

GDBusInterfaceInfo *foundry_ipc_command_line_service_interface_info (void);
guint foundry_ipc_command_line_service_override_properties (GObjectClass *klass, guint property_id_begin);


/* D-Bus method call completion functions: */
void foundry_ipc_command_line_service_complete_run (
    FoundryIpcCommandLineService *object,
    GDBusMethodInvocation *invocation,
    GUnixFDList *fd_list,
    gint exit_code);



/* D-Bus method calls: */
void foundry_ipc_command_line_service_call_run (
    FoundryIpcCommandLineService *proxy,
    const gchar *arg_directory,
    const gchar *const *arg_environment,
    const gchar *const *arg_argv,
    GVariant *arg_stdin_handle,
    GVariant *arg_stdout_handle,
    GVariant *arg_stderr_handle,
    const gchar *arg_object_path,
    GUnixFDList *fd_list,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean foundry_ipc_command_line_service_call_run_finish (
    FoundryIpcCommandLineService *proxy,
    gint* out_exit_code,
    GUnixFDList **out_fd_list,
    GAsyncResult *res,
    GError **error);

gboolean foundry_ipc_command_line_service_call_run_sync (
    FoundryIpcCommandLineService *proxy,
    const gchar *arg_directory,
    const gchar *const *arg_environment,
    const gchar *const *arg_argv,
    GVariant *arg_stdin_handle,
    GVariant *arg_stdout_handle,
    GVariant *arg_stderr_handle,
    const gchar *arg_object_path,
    GUnixFDList  *fd_list,
    gint* out_exit_code,
    GUnixFDList **out_fd_list,
    GCancellable *cancellable,
    GError **error);



/* ---- */

#define FOUNDRY_IPC_TYPE_COMMAND_LINE_SERVICE_PROXY (foundry_ipc_command_line_service_proxy_get_type ())
#define FOUNDRY_IPC_COMMAND_LINE_SERVICE_PROXY(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), FOUNDRY_IPC_TYPE_COMMAND_LINE_SERVICE_PROXY, FoundryIpcCommandLineServiceProxy))
#define FOUNDRY_IPC_COMMAND_LINE_SERVICE_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), FOUNDRY_IPC_TYPE_COMMAND_LINE_SERVICE_PROXY, FoundryIpcCommandLineServiceProxyClass))
#define FOUNDRY_IPC_COMMAND_LINE_SERVICE_PROXY_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), FOUNDRY_IPC_TYPE_COMMAND_LINE_SERVICE_PROXY, FoundryIpcCommandLineServiceProxyClass))
#define FOUNDRY_IPC_IS_COMMAND_LINE_SERVICE_PROXY(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), FOUNDRY_IPC_TYPE_COMMAND_LINE_SERVICE_PROXY))
#define FOUNDRY_IPC_IS_COMMAND_LINE_SERVICE_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), FOUNDRY_IPC_TYPE_COMMAND_LINE_SERVICE_PROXY))

typedef struct _FoundryIpcCommandLineServiceProxy FoundryIpcCommandLineServiceProxy;
typedef struct _FoundryIpcCommandLineServiceProxyClass FoundryIpcCommandLineServiceProxyClass;
typedef struct _FoundryIpcCommandLineServiceProxyPrivate FoundryIpcCommandLineServiceProxyPrivate;

struct _FoundryIpcCommandLineServiceProxy
{
  /*< private >*/
  GDBusProxy parent_instance;
  FoundryIpcCommandLineServiceProxyPrivate *priv;
};

struct _FoundryIpcCommandLineServiceProxyClass
{
  GDBusProxyClass parent_class;
};

GType foundry_ipc_command_line_service_proxy_get_type (void) G_GNUC_CONST;

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (FoundryIpcCommandLineServiceProxy, g_object_unref)
#endif

void foundry_ipc_command_line_service_proxy_new (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
FoundryIpcCommandLineService *foundry_ipc_command_line_service_proxy_new_finish (
    GAsyncResult        *res,
    GError             **error);
FoundryIpcCommandLineService *foundry_ipc_command_line_service_proxy_new_sync (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);

void foundry_ipc_command_line_service_proxy_new_for_bus (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
FoundryIpcCommandLineService *foundry_ipc_command_line_service_proxy_new_for_bus_finish (
    GAsyncResult        *res,
    GError             **error);
FoundryIpcCommandLineService *foundry_ipc_command_line_service_proxy_new_for_bus_sync (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);


/* ---- */

#define FOUNDRY_IPC_TYPE_COMMAND_LINE_SERVICE_SKELETON (foundry_ipc_command_line_service_skeleton_get_type ())
#define FOUNDRY_IPC_COMMAND_LINE_SERVICE_SKELETON(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), FOUNDRY_IPC_TYPE_COMMAND_LINE_SERVICE_SKELETON, FoundryIpcCommandLineServiceSkeleton))
#define FOUNDRY_IPC_COMMAND_LINE_SERVICE_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), FOUNDRY_IPC_TYPE_COMMAND_LINE_SERVICE_SKELETON, FoundryIpcCommandLineServiceSkeletonClass))
#define FOUNDRY_IPC_COMMAND_LINE_SERVICE_SKELETON_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), FOUNDRY_IPC_TYPE_COMMAND_LINE_SERVICE_SKELETON, FoundryIpcCommandLineServiceSkeletonClass))
#define FOUNDRY_IPC_IS_COMMAND_LINE_SERVICE_SKELETON(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), FOUNDRY_IPC_TYPE_COMMAND_LINE_SERVICE_SKELETON))
#define FOUNDRY_IPC_IS_COMMAND_LINE_SERVICE_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), FOUNDRY_IPC_TYPE_COMMAND_LINE_SERVICE_SKELETON))

typedef struct _FoundryIpcCommandLineServiceSkeleton FoundryIpcCommandLineServiceSkeleton;
typedef struct _FoundryIpcCommandLineServiceSkeletonClass FoundryIpcCommandLineServiceSkeletonClass;
typedef struct _FoundryIpcCommandLineServiceSkeletonPrivate FoundryIpcCommandLineServiceSkeletonPrivate;

struct _FoundryIpcCommandLineServiceSkeleton
{
  /*< private >*/
  GDBusInterfaceSkeleton parent_instance;
  FoundryIpcCommandLineServiceSkeletonPrivate *priv;
};

struct _FoundryIpcCommandLineServiceSkeletonClass
{
  GDBusInterfaceSkeletonClass parent_class;
};

GType foundry_ipc_command_line_service_skeleton_get_type (void) G_GNUC_CONST;

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (FoundryIpcCommandLineServiceSkeleton, g_object_unref)
#endif

FoundryIpcCommandLineService *foundry_ipc_command_line_service_skeleton_new (void);


/* ------------------------------------------------------------------------ */
/* Declarations for org.gnome.foundry.CommandLine */

#define FOUNDRY_IPC_TYPE_COMMAND_LINE (foundry_ipc_command_line_get_type ())
#define FOUNDRY_IPC_COMMAND_LINE(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), FOUNDRY_IPC_TYPE_COMMAND_LINE, FoundryIpcCommandLine))
#define FOUNDRY_IPC_IS_COMMAND_LINE(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), FOUNDRY_IPC_TYPE_COMMAND_LINE))
#define FOUNDRY_IPC_COMMAND_LINE_GET_IFACE(o) (G_TYPE_INSTANCE_GET_INTERFACE ((o), FOUNDRY_IPC_TYPE_COMMAND_LINE, FoundryIpcCommandLineIface))

struct _FoundryIpcCommandLine;
typedef struct _FoundryIpcCommandLine FoundryIpcCommandLine;
typedef struct _FoundryIpcCommandLineIface FoundryIpcCommandLineIface;

struct _FoundryIpcCommandLineIface
{
  GTypeInterface parent_iface;

  gboolean (*handle_open) (
    FoundryIpcCommandLine *object,
    GDBusMethodInvocation *invocation,
    GUnixFDList *fd_list,
    gint arg_fdnum);

};

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (FoundryIpcCommandLine, g_object_unref)
#endif

GType foundry_ipc_command_line_get_type (void) G_GNUC_CONST;

GDBusInterfaceInfo *foundry_ipc_command_line_interface_info (void);
guint foundry_ipc_command_line_override_properties (GObjectClass *klass, guint property_id_begin);


/* D-Bus method call completion functions: */
void foundry_ipc_command_line_complete_open (
    FoundryIpcCommandLine *object,
    GDBusMethodInvocation *invocation,
    GUnixFDList *fd_list,
    GVariant *fd);



/* D-Bus method calls: */
void foundry_ipc_command_line_call_open (
    FoundryIpcCommandLine *proxy,
    gint arg_fdnum,
    GUnixFDList *fd_list,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean foundry_ipc_command_line_call_open_finish (
    FoundryIpcCommandLine *proxy,
    GVariant ** out_fd,
    GUnixFDList **out_fd_list,
    GAsyncResult *res,
    GError **error);

gboolean foundry_ipc_command_line_call_open_sync (
    FoundryIpcCommandLine *proxy,
    gint arg_fdnum,
    GUnixFDList  *fd_list,
    GVariant ** out_fd,
    GUnixFDList **out_fd_list,
    GCancellable *cancellable,
    GError **error);



/* ---- */

#define FOUNDRY_IPC_TYPE_COMMAND_LINE_PROXY (foundry_ipc_command_line_proxy_get_type ())
#define FOUNDRY_IPC_COMMAND_LINE_PROXY(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), FOUNDRY_IPC_TYPE_COMMAND_LINE_PROXY, FoundryIpcCommandLineProxy))
#define FOUNDRY_IPC_COMMAND_LINE_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), FOUNDRY_IPC_TYPE_COMMAND_LINE_PROXY, FoundryIpcCommandLineProxyClass))
#define FOUNDRY_IPC_COMMAND_LINE_PROXY_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), FOUNDRY_IPC_TYPE_COMMAND_LINE_PROXY, FoundryIpcCommandLineProxyClass))
#define FOUNDRY_IPC_IS_COMMAND_LINE_PROXY(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), FOUNDRY_IPC_TYPE_COMMAND_LINE_PROXY))
#define FOUNDRY_IPC_IS_COMMAND_LINE_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), FOUNDRY_IPC_TYPE_COMMAND_LINE_PROXY))

typedef struct _FoundryIpcCommandLineProxy FoundryIpcCommandLineProxy;
typedef struct _FoundryIpcCommandLineProxyClass FoundryIpcCommandLineProxyClass;
typedef struct _FoundryIpcCommandLineProxyPrivate FoundryIpcCommandLineProxyPrivate;

struct _FoundryIpcCommandLineProxy
{
  /*< private >*/
  GDBusProxy parent_instance;
  FoundryIpcCommandLineProxyPrivate *priv;
};

struct _FoundryIpcCommandLineProxyClass
{
  GDBusProxyClass parent_class;
};

GType foundry_ipc_command_line_proxy_get_type (void) G_GNUC_CONST;

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (FoundryIpcCommandLineProxy, g_object_unref)
#endif

void foundry_ipc_command_line_proxy_new (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
FoundryIpcCommandLine *foundry_ipc_command_line_proxy_new_finish (
    GAsyncResult        *res,
    GError             **error);
FoundryIpcCommandLine *foundry_ipc_command_line_proxy_new_sync (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);

void foundry_ipc_command_line_proxy_new_for_bus (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
FoundryIpcCommandLine *foundry_ipc_command_line_proxy_new_for_bus_finish (
    GAsyncResult        *res,
    GError             **error);
FoundryIpcCommandLine *foundry_ipc_command_line_proxy_new_for_bus_sync (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);


/* ---- */

#define FOUNDRY_IPC_TYPE_COMMAND_LINE_SKELETON (foundry_ipc_command_line_skeleton_get_type ())
#define FOUNDRY_IPC_COMMAND_LINE_SKELETON(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), FOUNDRY_IPC_TYPE_COMMAND_LINE_SKELETON, FoundryIpcCommandLineSkeleton))
#define FOUNDRY_IPC_COMMAND_LINE_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), FOUNDRY_IPC_TYPE_COMMAND_LINE_SKELETON, FoundryIpcCommandLineSkeletonClass))
#define FOUNDRY_IPC_COMMAND_LINE_SKELETON_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), FOUNDRY_IPC_TYPE_COMMAND_LINE_SKELETON, FoundryIpcCommandLineSkeletonClass))
#define FOUNDRY_IPC_IS_COMMAND_LINE_SKELETON(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), FOUNDRY_IPC_TYPE_COMMAND_LINE_SKELETON))
#define FOUNDRY_IPC_IS_COMMAND_LINE_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), FOUNDRY_IPC_TYPE_COMMAND_LINE_SKELETON))

typedef struct _FoundryIpcCommandLineSkeleton FoundryIpcCommandLineSkeleton;
typedef struct _FoundryIpcCommandLineSkeletonClass FoundryIpcCommandLineSkeletonClass;
typedef struct _FoundryIpcCommandLineSkeletonPrivate FoundryIpcCommandLineSkeletonPrivate;

struct _FoundryIpcCommandLineSkeleton
{
  /*< private >*/
  GDBusInterfaceSkeleton parent_instance;
  FoundryIpcCommandLineSkeletonPrivate *priv;
};

struct _FoundryIpcCommandLineSkeletonClass
{
  GDBusInterfaceSkeletonClass parent_class;
};

GType foundry_ipc_command_line_skeleton_get_type (void) G_GNUC_CONST;

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (FoundryIpcCommandLineSkeleton, g_object_unref)
#endif

FoundryIpcCommandLine *foundry_ipc_command_line_skeleton_new (void);


G_END_DECLS

#endif /* __FOUNDRY_IPC_H__ */
